#include "dronedge.h"

static JSClassID js_drone_class_id;

static UA_Boolean pubsub_running = true;

static UA_UInt32 nbDrone;
static JSValueConst *drone_object_id_list;

static UA_UInt32 lastCheckpoint = 0;

static void js_drone_finalizer(JSRuntime *rt, JSValue val)
{
    JSDroneData *s = (JSDroneData *) JS_GetOpaque(val, js_drone_class_id);
    js_free_rt(rt, s);
}

static JSValue js_drone_ctor(JSContext *ctx, JSValueConst new_target, 
                             int argc, JSValueConst *argv)
{
    JSDroneData *s;
    JSValue obj = JS_UNDEFINED;
    JSValue proto;
    uint32_t uint32;
    
    s = (JSDroneData *) js_mallocz(ctx, sizeof(*s));
    if (!s)
        return JS_EXCEPTION;
    if (JS_ToUint32(ctx, &uint32, argv[0]))
        goto fail;
    s->id = (uint16_t)uint32;

    proto = JS_GetPropertyStr(ctx, new_target, "prototype");
    if (JS_IsException(proto))
        goto fail;
    obj = JS_NewObjectProtoClass(ctx, proto, js_drone_class_id);
    JS_FreeValue(ctx, proto);
    if (JS_IsException(obj))
        goto fail;
    JS_SetOpaque(obj, s);
    return obj;
 fail:
    js_free(ctx, s);
    JS_FreeValue(ctx, obj);
    return JS_EXCEPTION;
}

static JSValue js_drone_init(JSContext *ctx, JSValueConst this_val,
                             int argc, JSValueConst *argv)
{
    int nb;
    JSDroneData *s = (JSDroneData *) JS_GetOpaque2(ctx, this_val, js_drone_class_id);
    if (!s)
        return JS_EXCEPTION;
    if (JS_ToInt32(ctx, &nb, argv[0]))
        return JS_EXCEPTION;
    drone_object_id_list[nb] = this_val;
    return JS_UNDEFINED;
}

static JSValue js_drone_get(JSContext *ctx, JSValueConst this_val, int magic)
{
    JSDroneData *s = (JSDroneData *) JS_GetOpaque2(ctx, this_val, js_drone_class_id);
    if (!s)
        return JS_EXCEPTION;
    switch(magic) {
        case 0:
            return JS_NewUint32(ctx, s->id);
        case 1:
            return JS_NewFloat64(ctx, s->latitude);
        case 2:
            return JS_NewFloat64(ctx, s->longitude);
        case 3:
            return JS_NewFloat64(ctx, s->altitudeAbs);
        case 4:
            return JS_NewFloat64(ctx, s->altitudeRel);
        case 5:
            return JS_NewUint32(ctx, s->lastCheckpoint);
        default:
            return JS_EXCEPTION;
    }
}

static JSValue js_drone_set_checkpoint(JSContext *ctx, JSValueConst this_val,
                                       int argc, JSValueConst *argv)
{
    uint32_t uint32;
    if (JS_ToUint32(ctx, &uint32, argv[0]))
        return JS_EXCEPTION;
    lastCheckpoint = (uint16_t)uint32;
    return JS_UNDEFINED;
}

UA_UInt32 getLastCheckpoint(void)
{
    return lastCheckpoint;
}

static JSClassDef js_drone_class = {
    "Drone",
    .finalizer = js_drone_finalizer,
};

static const JSCFunctionListEntry js_drone_proto_funcs[] = {
    JS_CGETSET_MAGIC_DEF("id", js_drone_get, NULL, 0),
    JS_CGETSET_MAGIC_DEF("latitude", js_drone_get, NULL, 1),
    JS_CGETSET_MAGIC_DEF("longitude", js_drone_get, NULL, 2),
    JS_CGETSET_MAGIC_DEF("altitudeAbs", js_drone_get, NULL, 3),
    JS_CGETSET_MAGIC_DEF("altitudeRel", js_drone_get, NULL, 4),
    JS_CGETSET_MAGIC_DEF("lastCheckpoint", js_drone_get, NULL, 5),
    JS_CFUNC_DEF("init", 1, js_drone_init),
};

UA_UInt16 get_drone_id(UA_UInt32 nb) {
    JSDroneData *s = (JSDroneData *) JS_GetOpaque(drone_object_id_list[nb], js_drone_class_id);
    return s->id;
}

void stop_pubsub(void) {
    pubsub_running = false;
    free(drone_object_id_list);
}

VariableData pubsub_get_value(UA_String identifier) {
    VariableData var_details;
    UA_String name;

    for(UA_UInt32 i = 0; i < countof(droneVariableArray); i++) {
        name = UA_STRING(droneVariableArray[i].name);
        if(UA_String_equal(&identifier, &name)) {
            var_details = droneVariableArray[i];

            switch(var_details.type) {
                case UA_TYPES_DOUBLE:
                    *(UA_Double*)var_details.value = droneVariableArray[i].getter.getDouble();
                    break;
                case UA_TYPES_FLOAT:
                    *(UA_Float*)var_details.value = droneVariableArray[i].getter.getFloat();
                    break;
                case UA_TYPES_UINT32:
                    *(UA_UInt32*)var_details.value = droneVariableArray[i].getter.getUint32();
                    break;
                case UA_TYPES_DATETIME:
                    *(UA_DateTime*)var_details.value = droneVariableArray[i].getter.getTimestamp();
                    break;
                default:
                    UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "UA_TYPE not handled");
                    break;
            }
            break;
        }
    }

    return var_details;
}

void init_node_id(UA_UInt32 id, UA_UInt32 nb, UA_UInt32 magic) {
    JSDroneData *s = (JSDroneData *) JS_GetOpaque(drone_object_id_list[nb], js_drone_class_id);
    switch(magic) {
        case 0:
            s->latitudeId = id;
            break;
        case 1:
            s->longitudeId = id;
            break;
        case 2:
            s->altitudeAbsId = id;
            break;
        case 3:
            s->altitudeRelId = id;
            break;
        case 4:
            s->lastCheckpointId = id;
            break;
        default:
            break;
    }
}

void pubsub_update_coordinates(UA_UInt32 id, const UA_DataValue *var)
{
    JSDroneData *s;
    for(UA_UInt32 i = 0; i < nbDrone; i++) {
        s = (JSDroneData *) JS_GetOpaque(drone_object_id_list[i], js_drone_class_id);
        if (s->latitudeId == id) {
            s->latitude = *(UA_Double*) var->value.data;
            return;
        } else if (s->longitudeId == id) {
            s->longitude = *(UA_Double*) var->value.data;
            return;
        } else if (s->altitudeAbsId == id) {
            s->altitudeAbs = *(UA_Float*) var->value.data;
            return;
        } else if (s->altitudeRelId == id) {
            s->altitudeRel = *(UA_Float*) var->value.data;
            return;
        } else if (s->lastCheckpointId == id) {
            s->lastCheckpoint = *(UA_UInt32*) var->value.data;
            return;
        }
    }
    UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "NodeId not found");
}

void pubsub_print_coordinates(UA_UInt32 id, const UA_DataValue *var)
{
    JSDroneData *s;
    for(UA_UInt32 i = 0; i < nbDrone; i++) {
        s = (JSDroneData *) JS_GetOpaque(drone_object_id_list[i], js_drone_class_id);
        if (s->latitudeId == id) {
            s->latitude = *(UA_Double*) var->value.data;
            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "Received latitude of drone %d: %f°", s->id, s->latitude);
            return;
        } else if (s->longitudeId == id) {
            s->longitude = *(UA_Double*) var->value.data;
            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "Received longitude of drone %d: %f°", s->id, s->longitude);
            return;
        } else if (s->altitudeAbsId == id) {
            s->altitudeAbs = *(UA_Float*) var->value.data;
            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "Received absolute altitude of drone %d: %fm", s->id, s->altitudeAbs);
            return;
        } else if (s->altitudeRelId == id) {
            s->altitudeRel = *(UA_Float*) var->value.data;
            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "Received relative altitude of drone %d: %fm", s->id, s->altitudeRel);
            return;
        } else if (s->lastCheckpointId == id) {
            s->lastCheckpoint = *(UA_UInt32*) var->value.data;
            UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "Received checkpoint of drone %d: %dm", s->id, s->lastCheckpoint);
            return;
        }
    }
    UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "NodeId not found");
}

/*
 * arg 0 (string): IPv6 of the multicast group
 * arg 1 (string): port used for multicast communication
 * arg 2 (string): network interface used for multicast communication
 * arg 3 (int): ID of the drone
 * arg 4 (bool): true if there will be data to publish
*/
static JSValue js_run_pubsub(JSContext *ctx, JSValueConst this_val,
                             int argc, JSValueConst *argv)
{
    const char *ipv6;
    const char *port;
    const char *net_iface;
    char *not_const_net_iface;
    char urlBuffer[44];
    UA_UInt32 id;
    int res;

    ipv6 = JS_ToCString(ctx, argv[0]);
    port = JS_ToCString(ctx, argv[1]);
    net_iface = JS_ToCString(ctx, argv[2]);
    UA_snprintf(urlBuffer, sizeof(urlBuffer), "opc.udp://[%s]:%s/", ipv6, port);

    UA_String transportProfile =
        UA_STRING("http://opcfoundation.org/UA-Profile/Transport/pubsub-udp-uadp");

    UA_NetworkAddressUrlDataType networkAddressUrl;
    not_const_net_iface = strdup(net_iface);
    networkAddressUrl.networkInterface = UA_STRING(not_const_net_iface);
    networkAddressUrl.url              = UA_STRING(urlBuffer);

    if (JS_ToUint32(ctx, &id, argv[3]))
        return JS_EXCEPTION;

    if (JS_ToBool(ctx, argv[4])) {
        res = runPubsub(&transportProfile, &networkAddressUrl,
                        droneVariableArray, countof(droneVariableArray), id,
                        nbDrone, init_node_id, get_drone_id, pubsub_get_value,
                        pubsub_update_coordinates, &pubsub_running);
    } else {
        res = subscribeOnly(&transportProfile, &networkAddressUrl,
                            droneVariableArray, countof(droneVariableArray), id,
                            nbDrone, init_node_id, get_drone_id,
                            pubsub_print_coordinates, &pubsub_running);
    }

    JS_FreeCString(ctx, ipv6);
    JS_FreeCString(ctx, port);
    free(not_const_net_iface);
    JS_FreeCString(ctx, net_iface);
    return JS_NewInt32(ctx, res);
}

static JSValue js_init_pubsub(JSContext *ctx, JSValueConst this_val,
                              int argc, JSValueConst *argv)
{
    if (JS_ToUint32(ctx, &nbDrone, argv[0]))
        return JS_EXCEPTION;

    drone_object_id_list = (JSValue *) malloc(nbDrone * sizeof(JSValueConst));
    return JS_NewInt32(ctx, 0);
}

static JSValue js_stop_pubsub(JSContext *ctx, JSValueConst this_val,
                              int argc, JSValueConst *argv)
{
    pubsub_running = false;
    free(drone_object_id_list);
    return JS_NewInt32(ctx, 0);
}

static JSValue js_mavsdk_start(JSContext *ctx, JSValueConst this_val,
                               int argc, JSValueConst *argv)
{
    const char *url;
    const char *log_file;
    int timeout;
    int res;

    url = JS_ToCString(ctx, argv[0]);
    log_file = JS_ToCString(ctx, argv[1]);
    if (JS_ToInt32(ctx, &timeout, argv[2]))
        return JS_EXCEPTION;

    res = mavsdk_start(url, log_file, timeout);
    JS_FreeCString(ctx, url);
    JS_FreeCString(ctx, log_file);

    return JS_NewInt32(ctx, res);
}

static JSValue js_mavsdk_stop(JSContext *ctx, JSValueConst this_val,
                              int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_stop());
}

static JSValue js_mavsdk_reboot(JSContext *ctx, JSValueConst this_val,
                                int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_reboot());
}

static JSValue js_mavsdk_healthAllOk(JSContext *ctx, JSValueConst this_val,
                                     int argc, JSValueConst *argv)
{
    return JS_NewBool(ctx, mavsdk_healthAllOk());
}

static JSValue js_mavsdk_landed(JSContext *ctx, JSValueConst this_val,
                                int argc, JSValueConst *argv)
{
    return JS_NewBool(ctx, mavsdk_landed());
}

static JSValue js_mavsdk_isInManualMode(JSContext *ctx, JSValueConst thisVal,
                                        int argc, JSValueConst *argv)
{
    return JS_NewBool(ctx, mavsdk_isInManualMode());
}

static JSValue js_mavsdk_arm(JSContext *ctx, JSValueConst this_val,
                             int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_arm());
}

static JSValue js_mavsdk_setTargetCoordinates(JSContext *ctx,
                                              JSValueConst this_val,
                                              int argc, JSValueConst *argv)
{
    double la_arg_double;
    double lo_arg_double;
    double a_arg_double;

    if (JS_ToFloat64(ctx, &la_arg_double, argv[0]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &lo_arg_double, argv[1]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &a_arg_double, argv[2]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, mavsdk_setTargetCoordinates(la_arg_double,
                                                        lo_arg_double,
                                                        (float)a_arg_double));
}

static JSValue js_mavsdk_setManualControlInput(JSContext *ctx,
                                               JSValueConst thisVal,
                                               int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_setManualControlInput());
}

static JSValue js_mavsdk_setAltitude(JSContext *ctx, JSValueConst this_val,
                                     int argc, JSValueConst *argv)
{
    double altitude;

    if (JS_ToFloat64(ctx, &altitude, argv[0]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, mavsdk_setAltitude((float)altitude));
}

static JSValue js_mavsdk_setAirspeed(JSContext *ctx, JSValueConst this_val,
                                     int argc, JSValueConst *argv)
{
    double altitude;

    if (JS_ToFloat64(ctx, &altitude, argv[0]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, mavsdk_setAirspeed((float)altitude));
}

static JSValue js_mavsdk_getYaw(JSContext *ctx, JSValueConst this_val,
                                int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getYaw());
}

static JSValue js_mavsdk_getInitialLatitude(JSContext *ctx,
                                            JSValueConst this_val,
					                                  int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getInitialLatitude());
}

static JSValue js_mavsdk_getLatitude(JSContext *ctx, JSValueConst this_val,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getLatitude());
}

static JSValue js_mavsdk_getInitialLongitude(JSContext *ctx,
                                             JSValueConst this_val,
					                                   int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getInitialLongitude());
}

static JSValue js_mavsdk_getLongitude(JSContext *ctx, JSValueConst this_val,
                                      int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getLongitude());
}

static JSValue js_mavsdk_getTakeOffAltitude(JSContext *ctx,
                                            JSValueConst this_val,
					                                  int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getTakeOffAltitude());
}

static JSValue js_mavsdk_getInitialAltitude(JSContext *ctx,
                                            JSValueConst this_val,
                                            int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getInitialAltitude());
}

static JSValue js_mavsdk_getAltitude(JSContext *ctx, JSValueConst this_val,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getAltitude());
}

static JSValue js_mavsdk_getAltitudeRel(JSContext *ctx, JSValueConst this_val,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, mavsdk_getAltitudeRel());
}

static JSValue js_mavsdk_loiter(JSContext *ctx, JSValueConst this_val,
                                int argc, JSValueConst *argv)
{
    double radius = 100;
    if (argc > 0) {
        if (JS_ToFloat64(ctx, &radius, argv[0]))
            return JS_EXCEPTION;
    } 

    return JS_NewInt32(ctx, mavsdk_loiter((float)radius));
}

static JSValue js_mavsdk_triggerParachute(JSContext *ctx, JSValueConst this_val,
                                     int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_triggerParachute());
}

static JSValue js_mavsdk_takeOff(JSContext *ctx, JSValueConst this_val,
                                 int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_takeOff());
}

static JSValue js_mavsdk_takeOffAndWait(JSContext *ctx, JSValueConst this_val,
                                 int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_takeOffAndWait());
}

static JSValue js_mavsdk_land(JSContext *ctx, JSValueConst this_val,
                              int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, mavsdk_land());
}

static const JSCFunctionListEntry js_mavsdk_funcs[] = {
    JS_CFUNC_DEF("start", 3, js_mavsdk_start ),
    JS_CFUNC_DEF("stop", 0, js_mavsdk_stop ),
    JS_CFUNC_DEF("reboot", 0, js_mavsdk_reboot ),
    JS_CFUNC_DEF("healthAllOk", 0, js_mavsdk_healthAllOk ),
    JS_CFUNC_DEF("landed", 0, js_mavsdk_landed ),
    JS_CFUNC_DEF("isInManualMode", 0, js_mavsdk_isInManualMode ),
    JS_CFUNC_DEF("arm", 0, js_mavsdk_arm ),
    JS_CFUNC_DEF("setTargetCoordinates", 3, js_mavsdk_setTargetCoordinates ),
    JS_CFUNC_DEF("setManualControlInput", 0, js_mavsdk_setManualControlInput ),
    JS_CFUNC_DEF("setAltitude", 1, js_mavsdk_setAltitude ),
    JS_CFUNC_DEF("setAirspeed", 1, js_mavsdk_setAirspeed ),
    JS_CFUNC_DEF("loiter", 0, js_mavsdk_loiter ),
    JS_CFUNC_DEF("triggerParachute", 0, js_mavsdk_triggerParachute ),
    JS_CFUNC_DEF("getYaw", 0, js_mavsdk_getYaw ),
    JS_CFUNC_DEF("getInitialLatitude", 0, js_mavsdk_getInitialLatitude ),
    JS_CFUNC_DEF("getLatitude", 0, js_mavsdk_getLatitude ),
    JS_CFUNC_DEF("getInitialLongitude", 0, js_mavsdk_getInitialLongitude ),
    JS_CFUNC_DEF("getLongitude", 0, js_mavsdk_getLongitude ),
    JS_CFUNC_DEF("getTakeOffAltitude", 0, js_mavsdk_getTakeOffAltitude ),
    JS_CFUNC_DEF("getInitialAltitude", 0, js_mavsdk_getInitialAltitude ),
    JS_CFUNC_DEF("getAltitude", 0, js_mavsdk_getAltitude ),
    JS_CFUNC_DEF("getAltitudeRel", 0, js_mavsdk_getAltitudeRel ),
    JS_CFUNC_DEF("takeOff", 0, js_mavsdk_takeOff ),
    JS_CFUNC_DEF("takeOffAndWait", 0, js_mavsdk_takeOffAndWait ),
    JS_CFUNC_DEF("land", 0, js_mavsdk_land ),
    JS_CFUNC_DEF("initPubsub", 1, js_init_pubsub ),
    JS_CFUNC_DEF("runPubsub", 5, js_run_pubsub ),
    JS_CFUNC_DEF("setCheckpoint", 1, js_drone_set_checkpoint ),
    JS_CFUNC_DEF("stopPubsub", 0, js_stop_pubsub ),
};

static int js_mavsdk_init(JSContext *ctx, JSModuleDef *m)
{
    JSValue drone_proto, drone_class;

    JS_NewClassID(&js_drone_class_id);
    JS_NewClass(JS_GetRuntime(ctx), js_drone_class_id, &js_drone_class);

    drone_proto = JS_NewObject(ctx);
    JS_SetPropertyFunctionList(ctx, drone_proto, js_drone_proto_funcs,
                               countof(js_drone_proto_funcs));

    drone_class = JS_NewCFunction2(ctx, js_drone_ctor, "Drone", 1,
                                   JS_CFUNC_constructor, 0);
    JS_SetConstructor(ctx, drone_class, drone_proto);
    JS_SetClassProto(ctx, js_drone_class_id, drone_proto);

    JS_SetModuleExport(ctx, m, "Drone", drone_class);

    return JS_SetModuleExportList(ctx, m, js_mavsdk_funcs,
                                  countof(js_mavsdk_funcs));
}

JSModuleDef *js_init_module(JSContext *ctx, const char *module_name)
{
    JSModuleDef *m;
    m = JS_NewCModule(ctx, module_name, js_mavsdk_init);
    if (!m)
        return NULL;
    JS_AddModuleExportList(ctx, m, js_mavsdk_funcs, countof(js_mavsdk_funcs));
    JS_AddModuleExport(ctx, m, "Drone");
    return m;
}
